/*
 * ntpdate.c: Program to obtain the UTC time from a time server,
 * and to print out a localtime strings which is suitable for use by settime.
 *
 * NOTE: We actually don't use NTP. The name is simply to keep the TiVo
 * happy.
 *
 * Copyright (c) 2004 Warren Toomey and Andrew Tridgell.
 * The program is placed under the GPL.
 */

#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tivotimelib.h"

const char *DaylightSavingsMarker = "/var/tmp/dst_lastmodified";
static const char *Version= "$Revision: 1.15 $";
static int set_the_time=1;		/* We want to set the time */
extern int verbose;


/* Open a socket to a tcp remote host with the specified port */
static int open_socket_out(const char *host, int port)
{
  struct sockaddr_in sock_out;
  int res;
  struct in_addr addr;

  if (verbose)
    fprintf(stderr, "Connecting to time server %s\n",host);
  res = socket(PF_INET, SOCK_STREAM, 0);
  if (res == -1) {
    if (verbose)
      fprintf(stderr, "Failed to obtain a socket: %s\n", strerror(errno));
    else
      syslog(LOG_INFO, "Failed to obtain a socket: %s\n", strerror(errno));
    return(-1);
  }

  inet_aton(host, &addr);
  sock_out.sin_addr = addr;
  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *) & sock_out, sizeof(sock_out)) != 0) {
    close(res);
    if (verbose)
      fprintf(stderr, "Failed to connect to %s: %s\n", host, strerror(errno));
    else
      syslog(LOG_INFO, "Failed to connect to %s: %s\n", host, strerror(errno));
    return(-1);
  }
  return(res);
}


static void 
tivo_time(int s, long ofs)
{
  struct timeval tv;
  unsigned long t;
  if (read(s, &t, sizeof(t)) != 4) {
    if (verbose)
      fprintf(stderr, "Unable to read time from server: %s\n",strerror(errno));
    else
      syslog(LOG_INFO, "Unable to read time from server: %s\n",strerror(errno));
    exit(1);
  }

  t = ntohl(t);
  t += (ofs - 2208988800U);
  if (set_the_time) {
    if (verbose)
      printf("Setting the local time to %s", asctime(gmtime(&t)));
    syslog(LOG_INFO, "Setting the local time to %s",
                                                asctime(gmtime(&t)));
    tv.tv_sec = t;
    tv.tv_usec = 0;
    if (settimeofday(&tv, NULL) == -1) {
      if (verbose)
        printf("Couldn't settimeofday\n");
      else
        syslog(LOG_INFO, "Couldn't settimeofday\n");
    }
  }
}

void do_daylight_saving_change(long prevoffset, long offset)
{
  FILE *zout;
  struct stat sb;
  time_t secs;
  struct timeval tv;
  int err;

  if (verbose)
    fprintf(stderr, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);
  else
    syslog(LOG_INFO, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);

  /* We have to be careful and not adjust the clock several times a la
   * Windoze. If the DaylightSavingsMarker is more than a day old, then
   * make the change. Otherwise, don't make the change.
   */
  secs = time(NULL);
  err=stat(DaylightSavingsMarker, &sb);
  if ((err==0) && ((secs - sb.st_mtime)<86400)) {
    if (verbose)
      fprintf(stderr, "Not doing daylight saving change: I already did it\n");
    else
      syslog(LOG_INFO, "Not doing daylight saving change: I already did it\n");
    exit(0);
  }

  /* Make the change, touch the file. */
  zout= fopen(DaylightSavingsMarker, "w");
  if (zout==NULL) exit(1);
  fprintf(zout, "Daylight saving change: offset %ld -> %ld in secs\n",
                                                        prevoffset,offset);
  fclose(zout);
  
  secs= secs + (offset - prevoffset);
  
  if (set_the_time) {
    if (verbose)
      printf("Setting the Local time to %s", asctime(gmtime(&secs)));
    syslog(LOG_INFO, "Setting the local time to %s",
                                                asctime(gmtime(&secs)));
    tv.tv_sec = secs;
    tv.tv_usec = 0;
    if (settimeofday(&tv, NULL) == -1) {
      if (verbose)
        printf("Couldn't settimeofday\n");
      else
        syslog(LOG_INFO, "Couldn't settimeofday\n");
    }
  }
  exit(0);
}

int 
main(int argc, char *argv[])
{
  char *host;
  int ch, sock;
  int servernum;
  long offset,prevoffset;

  if (argc < 2) {
    fprintf(stderr, "ntpdate [-B] [-b] [d] [-q] [-v] server [server ...]\n");
    exit(1);
  }

  openlog("ntpdate", LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "fake ntpdate has started");

  while ((ch = getopt(argc, argv, "Bbdqv")) != -1)
    switch (ch) {
      case 'B':
      case 'b': break;
      case 'd':
      case 'q': set_the_time=0; break;
      case 'v': verbose=1; break;
    }
  argc -= optind;
  argv += optind;

  offset = get_timezone_offset(0);
  if (offset==-1) {
    if (verbose) 
      printf("PANIC: couldn't find TiVo's time zone\n");
    syslog(LOG_INFO, "PANIC: couldn't find TiVo's time zone\n");
    exit(1);
  }

  for (servernum=0; servernum<argc; servernum++) {
    host = argv[servernum];
    sock = open_socket_out(host, 37);
    if (sock<0) continue;
    syslog(LOG_INFO, "Connected to time server %s", host);
    tivo_time(sock, offset);
    close(sock);
    closelog();
    exit(0);
  }

  /* We failed to connect to any servers */
  if (verbose)
    fprintf(stderr, "Failed to connect to any of the time servers\n");
  else
    syslog(LOG_INFO, "Failed to connect to any of the time servers\n");

  /* Do nothing if the offset now and 2 hours ago are the same */
  prevoffset = get_timezone_offset(2);
  if (offset==prevoffset) exit(1);

  do_daylight_saving_change(prevoffset,offset);
  exit(0);
}
